! ===============================================================================================================================
! MODULE       : albedo
!
! CONTACT      : orchidee-help _at_ ipsl.jussieu.fr
!
! LICENCE      : IPSL (2006)
! This software is governed by the CeCILL licence see ORCHIDEE/ORCHIDEE_CeCILL.LIC
!
!>\BRIEF        Calculate the surface albedo using a variety of different schemes
!!
!! \n DESCRIPTION : This module includes the mechanisms for
!! 1. :: albedo_two_stream computes the two_stream approach of Pinty et al 2006.
!!           Requires an effective leaf area index and depends on the solar angle.
!!           Separates light into NIR and VIS, and diffuse and direct illumination
!!
!! RECENT CHANGE(S): None
!!
!! REFERENCES(S)    : 
!!                    Chalita, S. and H Le Treut (1994), The albedo of temperate
!!                    and boreal forest and the Northern Hemisphere climate: a 
!!                    sensitivity experiment using the LMD GCM, 
!!                    Climate Dynamics, 10 231-240.
!!
!!                    B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, 
!!                    N. Gobron and M. M. Verstraete (2006). Simplifying the 
!!                    Interaction of Land Surfaces with Radiation for Relating 
!!                    Remote Sensing Products to Climate Models. Journal of 
!!                    Geophysical Research. Vol 111, D02116.
!!
!! SVN              :
!! $HeadURL: svn://forge.ipsl.jussieu.fr/orchidee/perso/matthew.mcgrath/TRUNK/ORCHIDEE/src_sechiba/albedo.f90 $
!! $Date: 2012-07-19 15:12:52 +0200 (Thu, 1 Nov 2012) $
!! $Revision: 947 $
!! \n
!_ ================================================================================================================================

MODULE albedo
  !
  USE ioipsl
  !
  ! modules used :
  USE constantes
  USE constantes_soil
  USE pft_parameters
  USE structures
  USE interpol_help
  USE ioipsl_para, ONLY : ipslerr_p
  USE stomate_laieff

  IMPLICIT NONE

  PRIVATE :: print_debugging_albedo_info,calculate_snow_albedo,&
       twostream_solver,gammas
  PUBLIC :: albedo_two_stream
                                                                        
CONTAINS

!! ==============================================================================================================================
!! SUBROUTINE   : albedo_two_stream
!!
!>\BRIEF        This subroutine calculates the albedo for two stream radiation transfer model.  This routine includes
!!              the effect of snow on the background albedo, through one of two methods.
!!
!! DESCRIPTION  : The albedo for two stream radiation transfer model  is calculated for both the visible and near-infrared 
!! domain. First the mean albedo of the bare soil is calculated. Two options exist: 
!! either the soil albedo depends on soil wetness (drysoil_frac variable), or the soil albedo 
!! is set to a mean soil albedo value.
!!
!!    NOTE: the main output variable, albedo, is an unweighted average of the direct and diffuse albedos
!!          for each grid point...this is done in this way right now to be consistent with the new scheme,
!!          but it doesn't have to be combined like that once we have an energy budget which can
!!          use both types of light directly
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S): ::albedo
!!
!! REFERENCE(S) : B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron and M. M. Verstraete (2006).
!! Simplifying the Interaction of Land Surfaces with Radiation for Relating Remote Sensing Products to Climate Models.
!! Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!! \n
!!
!!
!!
!_ ================================================================================================================================

  SUBROUTINE albedo_two_stream(npts, &
       drysoil_frac, veget_max, coszang, soilalb_dry, &
       soilalb_wet, frac_nobio, soilalb_moy, bckgrnd_alb, & 
       alb_bare, albedo, snow, snow_age, snow_nobio, &
       snow_nobio_age, albedo_snow, albedo_glob, &
       circ_class_biomass, circ_class_n, z0m, &
       Isotrop_Abs_Tot_p, Isotrop_Tran_Tot_p, laieff_fit, &
       albedo_pft, frac_snow_veg, frac_snow_nobio,&
       Collim_Abs_Tot_mean, Collim_Alb_Tot_mean, &
       laieff_collim, laieff_isotrop)

 !! 0. Variable and parameter declaration

    !! 0.1 Input variables
 
   INTEGER(i_std), INTENT(in)                                :: npts                   !! Domain size - Number of land pixels  (unitless)       
   REAL(r_std), DIMENSION(:), INTENT(in)                     :: drysoil_frac           !! Fraction of  dry soil (unitless)
   REAL(r_std), DIMENSION(:), INTENT(in)                     :: coszang                !! The cosine of the solar zenith angle (unitless)
   REAL(r_std), DIMENSION(:,:), INTENT(in)                   :: veget_max              !! PFT coverage fraction of a PFT (= ind*cn_ind) 
                                                                                       !! (m^2 m^{-2})
   REAL(r_std), DIMENSION(:,:), INTENT(in)                   :: soilalb_dry            !! Albedo values for the dry bare soil (unitless)
   REAL(r_std), DIMENSION(:,:), INTENT(in)                   :: soilalb_wet            !! Albedo values for the wet bare soil (unitless)
   REAL(r_std), DIMENSION(:,:), INTENT(in)                   :: soilalb_moy            !! Albedo values for the mean bare soil (unitless)
   REAL(r_std), DIMENSION(:,:), INTENT(in)                   :: bckgrnd_alb             !! Background albedo values if read in from map 
                                                                                       !! (see condveg_background_soilalb)
   REAL(r_std),DIMENSION (:), INTENT(in)                     :: snow                   !! Snow mass in vegetation (kg m^{-2})           
   REAL(r_std),DIMENSION (:,:), INTENT(in)                   :: snow_nobio             !! Snow mass on continental ice, lakes, etc. (kg m^{-2})      
   REAL(r_std),DIMENSION (:), INTENT(in)                     :: snow_age               !! Snow age (days)        
   REAL(r_std),DIMENSION (:,:), INTENT(in)                   :: snow_nobio_age         !! Snow age on continental ice, lakes, etc. (days)    
   REAL(r_std),DIMENSION (:,:), INTENT(in)                   :: frac_nobio             !! Fraction of non-vegetative surfaces, i.e. 
                                                                                       !! continental ice, lakes, etc. (unitless)     
   REAL(r_std), DIMENSION(:,:,:,:,:), INTENT(IN)             :: circ_class_biomass     !! Stem diameter 
                                                                                       !! @tex $(m)$ @endtex
   REAL(r_std), DIMENSION(:,:,:), INTENT(IN)                 :: circ_class_n           !! Number of trees within each circumference
   REAL(r_std), DIMENSION(:), INTENT(in)                     :: z0m                    !! Roughness height for momentum of vegetated part (m)
   TYPE(laieff_type),DIMENSION (:,:,:),INTENT(in)            :: laieff_fit             !! Fitted parameters for the effective LAI
   
   !! 0.2 Output variables
   
   REAL(r_std), DIMENSION(:,:), INTENT(out)                  :: alb_bare               !! Mean bare soil albedo for visible and near-infrared
   REAL(r_std), DIMENSION (:,:), INTENT (out)                :: albedo                 !! Albedo (two stream radiation transfer model)
                                                                                       !! for visible and near-infrared range 
                                                                                       !! (unitless)
   REAL(r_std),DIMENSION (:), INTENT (out)                   :: albedo_snow            !! Snow albedo (unitless ratio)     
   REAL(r_std),DIMENSION (:), INTENT (out)                   :: albedo_glob            !! Mean albedo (unitless ratio)
   REAL(r_std),DIMENSION (:,:,:), INTENT (out)               :: Isotrop_Abs_Tot_p      !! Absorbed radiation per layer for photosynthesis
   REAL(r_std),DIMENSION (:,:,:), INTENT (out)               :: Isotrop_Tran_Tot_p     !! Transmitted radiation per layer for photosynthesis
   REAL(r_std), DIMENSION (:,:,:), INTENT (out)              :: albedo_pft             !! Albedo (two stream radiation transfer model)
                                                                                       !! for visible and near-infrared range 
                                                                                       !! for each PFT (unitless)
   REAL(r_std), DIMENSION(:,:), INTENT(in)                   :: frac_snow_nobio        !! Fraction of snow on continental ice, lakes, etc. 
                                                                                       !! (unitless ratio)
   REAL(r_std),DIMENSION(:), INTENT(in)                      :: frac_snow_veg          !! The fraction of ground covered with snow, between 0 and 1
   REAL(r_std),DIMENSION(nlevels_tot), INTENT (out)          :: Collim_Abs_Tot_mean    !! collimated total absorption for a given level        * changed for enerbil integration
   REAL(r_std),DIMENSION(nlevels_tot), INTENT (out)          :: Collim_Alb_Tot_mean    !! Collimated (direct) total albedo for a given level   * changed for enerbil integration
   REAL(r_std), DIMENSION(nlevels_tot,npts,nvm), INTENT(out) :: laieff_collim          !! Leaf Area Index Effective for direct light
   REAL(r_std), DIMENSION(nlevels_tot,npts,nvm), INTENT(out) :: laieff_isotrop         !! Leaf Area Index Effective converts
                                                                                       !! 3D lai into 1D lai for two stream
                                                                                       !! radiation transfer model...this is for 
                                                                                       !! isotropic light and only calculated once per day
                                                                                       !! @tex $(m^{2} m^{-2})$ @endtex      



   !! 0.3 Modified variables
 
 
   !! 0.4 Local variables

   REAL(r_std),DIMENSION (nlevels_tot)   :: laieff_isotrop_pft, laieff_collim_pft   !! 
   REAL(r_std)                           :: cosine_sun_angle    !! the cosine of the solar zenith angle
   INTEGER(i_std)                        :: ks                  !! Index for visible and near-infraread range
   INTEGER(i_std)                        :: ivm                 !! Index for vegetative PFTs
   INTEGER(i_std)                        :: ipts                !! Index for spatial domain
   INTEGER(i_std)                        :: ilevel              !! Index for canopy levels
   REAL(r_std)                           :: leaf_reflectance
   REAL(r_std)                           :: leaf_transmittance
   REAL(r_std)                           :: br_base_temp,br_base_temp_collim,br_base_temp_isotrop
   REAL(r_std)                           :: leaf_psd_temp
   REAL(r_std)                           :: leaf_single_scattering_albedo

   REAL(r_std),DIMENSION(nlevels_tot)    :: Collim_Tran_Uncoll  !! collimated total transmission of uncollided light for a given level 
   REAL(r_std),DIMENSION(nlevels_tot)    :: Collim_Tran_Coll    !! collimated total transmission for a given level
   REAL(r_std),DIMENSION(nlevels_tot)    :: Collim_Abs_Tot      !! collimated total absorption for a given level        * changed for enerbil integration
   REAL(r_std),DIMENSION(nlevels_tot)    :: Collim_Alb_Tot      !! Collimated (direct) total albedo for a given level   * changed for enerbil integration

   REAL(r_std),DIMENSION(nlevels_tot)    :: Isotrop_Alb_Tot     !! isotropic (diffuse) total albedo for a given level
   REAL(r_std),DIMENSION(nlevels_tot)    :: Isotrop_Tran_Uncoll !! isotropic total transmission of uncollided light for a given level
   REAL(r_std),DIMENSION(nlevels_tot)    :: Isotrop_Tran_Coll   !! isotropic total transmission for a given level
   REAL(r_std),DIMENSION(nlevels_tot)    :: Isotrop_Abs_Tot     !! isotropic total absporbtion for a given level
   INTEGER :: jv, inno
   REAL(r_std) :: alb_nobio

   REAL(r_std), DIMENSION(npts,nvm)      :: frac_snow_veg_loc   !! the fraction of ground covered with snow, between 0 and 1



   REAL(r_std):: laieff_collim_1, laieff_isotrop_1, Collim_Alb_Tot_1,Collim_Tran_Tot_1,Collim_Abs_Tot_1,&
                 Isotrop_Alb_Tot_1,Isotrop_Tran_Tot_1,Isotrop_Abs_Tot_1,&
                 Collim_Tran_Uncollided_1, Isotrop_Tran_Uncollided_1  !! Values for the one level solution

   REAL(r_std):: converged_albedo
   REAL(r_std):: isotrop_angle
   LOGICAL :: lconverged
   LOGICAL :: lprint                                                  !! a flag for printing some debug statements
   REAL(r_std), DIMENSION(npts,nvm,n_spectralbands) :: snowa_veg      !! snow albedo due to vegetated surfaces, between 0 and 1
   REAL(r_std), DIMENSION(npts,nnobio,n_spectralbands) :: snowa_nobio !! Albedo of snow covered area on continental ice, 
                                                                      !! lakes, etc. (unitless ratio)  

!_ ================================================================================================================================

   IF (printlev>=2) WRITE(numout,*) 'Entering albedo_two_stream'

   !! Initialize local printlev
    !printlev_loc=get_printlev('albedo') 

   IF (printlev_loc>=3) WRITE(numout,*) 'nlevels_tot',nlevels_tot
   IF (printlev_loc>=4) THEN
      DO ivm = 1,nvm
         WRITE(numout,*) 'laieff_isotrop start albedo for ivm', &
              ivm,'is', laieff_isotrop(:,:,ivm)
      ENDDO
      WRITE(numout,*)'albedo.f90, coszang',coszang(:)
   ENDIF

   !! 1. We will now calculate the background reflectance to be used in the model.
   !  The parameters read in from the input file do not include the effect 
   !  of snow, so we need to figure out the snow's contribution for each 
   !  grid space
   !  Initialize some output variables
   DO ipts = 1, npts
      albedo_snow(ipts) = zero
   ENDDO

   Isotrop_Abs_Tot_p(:,:,:)=zero
   Isotrop_Tran_Tot_p(:,:,:)=un
   albedo_pft(:,:,:)=zero

   Collim_Abs_Tot_mean(:) = zero
   Collim_Alb_Tot_mean(:) = zero

   ! We need to calculate the LAI effective for this solar angle. The other LAI
   ! effective that we calculated was computed at an angle of 60.0 degrees for
   ! the isotropic contribution. Notice that, in many situations, the results 
   ! will be exactly the same.  One can show that Pgap is proportional to 
   ! 1/cos(theta) if the level bottom is below all canopy elements, which
   ! means that the cos(theta) in the calculation of the LAIeff will be
   ! cancelled out
   isotrop_angle=COS(60.0/180.0*pi)
   DO ipts=1,npts
      DO ivm=1,nvm
         DO ilevel=1,nlevels_tot
            laieff_isotrop(ilevel,ipts,ivm) = &
                 calculate_laieff_fit(isotrop_angle,laieff_fit(ipts,ivm,ilevel))
            laieff_collim(ilevel,ipts,ivm) = &
                 calculate_laieff_fit(coszang(ipts),laieff_fit(ipts,ivm,ilevel))
            ! Because we are fitting these values, it's possible that some 
            ! will drop below zero.  This causes a serious problem here,
            ! so let's make sure this doesn't happen.  Unfortunately, by
            ! doing this, we lose a test of things going wrong (an unfitted
            ! negative LAIeff can be the sign of another problem), but we
            ! don't really have a choice.
            IF(laieff_isotrop(ilevel,ipts,ivm) .LT. min_stomate) &
                 laieff_isotrop(ilevel,ipts,ivm)=zero
            IF(laieff_collim(ilevel,ipts,ivm) .LT. min_stomate) &
                 laieff_collim(ilevel,ipts,ivm)=zero
         ENDDO
      ENDDO
   ENDDO

   ! Debug
   IF(printlev_loc>=4)THEN
      DO ivm=1,nvm
         IF(ivm == test_pft)THEN
            DO ipts = 1, npts
               IF(ipts == test_grid)THEN
                  WRITE(numout,*) 'Laieff_collim and laieff_isotrop in albedo'
                  WRITE(numout,*) 'ACOS(coszang(ipts))/pi*180: ',&
                       ACOS(coszang(ipts))/pi*180.0_r_std
                  WRITE(numout,*) 'coszang(ipst)', coszang
                  DO ilevel=nlevels_tot,1,-1
                     WRITE(numout,*) ilevel,laieff_collim(ilevel,ipts,ivm),&
                          laieff_isotrop(ilevel,ipts,ivm)
                  END DO
                  DO ilevel=nlevels_tot,1,-1
                     WRITE(numout,*) 'Fitting parameters: ',&
                          laieff_fit(ipts,ivm,ilevel)%a,&
                          laieff_fit(ipts,ivm,ilevel)%b,&
                          laieff_fit(ipts,ivm,ilevel)%c,&
                          laieff_fit(ipts,ivm,ilevel)%d,&
                          laieff_fit(ipts,ivm,ilevel)%e
                  ENDDO
               ENDIF
            ENDDO
         ENDIF
      ENDDO
   ENDIF
   !-

   ! Calculate the snow albedo
   CALL calculate_snow_albedo(npts,  coszang,  snow,  snow_nobio, &
        snow_age, snow_nobio_age,  frac_nobio,  albedo_snow, &
        snowa_veg,  frac_snow_veg,  snowa_nobio,  frac_snow_nobio, &
        veget_max, z0m)

   ! Check for possible problems in the calcualtion of the fraction
   ! of the grid cell covered by snow. Fractions should be positive
   ! numbers
   DO ipts=1,npts
      IF(frac_snow_veg(ipts) .LT. 0.0)THEN
         WRITE(numout,*) 'SNOWFRAC ipts ',ipts,frac_snow_veg(ipts)
         CALL ipslerr_p (3,'albedo', 'snow frac error','','')
      END IF
     
      DO inno=1,nnobio
         IF(frac_snow_nobio(ipts,inno) .LT. 0.0)THEN
            WRITE(numout,*) 'SNOWFRAC ipts inno',ipts,inno,frac_snow_nobio(ipts,inno)
            CALL ipslerr_p (3,'albedo', 'nobio snow frac error','','')
         ENDIF
      ENDDO
   ENDDO

   ! Initialize the albedo in every square for both spectra by calculating the 
   ! bare soil albedo
   DO ipts = 1, npts
      DO ks = 1, n_spectralbands

         ! If 'alb_bare_model' is set to TRUE,  the soil albedo calculation 
         ! depends on soil moisture
         ! If 'alb_bare_model' is set to FALSE, the mean soil albedo is used 
         ! without the dependance on soil moisture
         ! see subroutine 'conveg_soilalb'
         IF ( alb_bare_model ) THEN
            alb_bare(ipts,ks) = soilalb_wet(ipts,ks) + drysoil_frac(ipts) * &
                 (soilalb_dry(ipts,ks) -  soilalb_wet(ipts,ks))
         ELSE
            alb_bare(ipts,ks) = soilalb_moy(ipts,ks)
         ENDIF

         ! we need to take into account any snow that has fallen on the bare 
         ! soil since there may be some bare soil, initialize the albedo with 
         ! this fraction
         albedo(ipts,ks) = veget_max(ipts,1) * (alb_bare(ipts,ks)*&
              (un-frac_snow_veg(ipts)) + &
              frac_snow_veg(ipts)*snowa_veg(ipts,1,ks))
      
      ENDDO ! ks = 1, n_spectralbands
   ENDDO ! ipts = 1, npts
 
   !! 2. Calculation of albedo of the whole grid cell, taking into account all 
   !!    PFTs (except bare soil) and spectral bands
   DO ipts = 1, npts ! loop over the grid squares

      !+++CHECK+++
      ! observations from the MERGE, sinang is calculated as cos of the zenith
      ! angle in solar.f90 (i.e. it is the cosine of the zenith angle). 
      ! Then maybe also the cosine_sun_angle is a redundant variable.
      cosine_sun_angle = coszang(ipts)
      !+++++++++++

      ! Debug
      IF(printlev_loc >= 4) THEN
         WRITE(numout,*)'cosine_sun_angle',cosine_sun_angle
      ENDIF

      ! For the coupled model, the angles can be negative.
      ! Offline, they are always between zero and 90 degrees.
      IF(cosine_sun_angle .LE. min_sechiba)THEN

         ! It's night, so no sunlight...don't need to calculate albedo.  
         ! Set it equal to zero, because the snow albedo has already been 
         ! calculated and it looks funny on the coupled run maps otherwise.
         albedo(ipts,:)=zero
         CYCLE

      ENDIF

      ! Are there any non-bio PFTs here that we need to take into account?
      DO ks = 1, n_spectralbands ! Loop over # of spectra
         DO jv = 1, nnobio 
            ! now update the albedo
            IF ( jv .EQ. iice ) THEN
               alb_nobio = alb_ice(ks)
            ELSE
               WRITE(numout,*) 'jv=',jv
               CALL ipslerr_p (3,'Albedo.f90', &
                    'DO NOT KNOW ALBEDO OF THIS SURFACE TYPE','','')
            ENDIF

            ! this takes into account both snow covered and non-snow 
            ! covered non-bio regios in all grid squares
            albedo(ipts,ks) = albedo(ipts,ks) + &
                 ( frac_nobio(ipts,jv) ) * &
                 ( (un-frac_snow_nobio(ipts,jv)) * alb_nobio + &
                 ( frac_snow_nobio(ipts,jv)  ) * snowa_nobio(ipts,jv,ks)   )

            albedo_snow(ipts) = albedo_snow(ipts) + &
                 ( frac_nobio(ipts,jv) ) * ( frac_snow_nobio(ipts,jv) ) * &
                 snowa_nobio(ipts,jv,ks)/REAL(n_spectralbands,r_std) 

         ENDDO ! jv = 1, nnobio 
      ENDDO ! ks = 1, n_spectralbands 
      

      DO ivm = 2, nvm  ! Loop over # of PFTs 

         ! Use local variable. Note that the dimension are changed
         ! a PFT-specific frac_snow_veg is now used. In the future
         ! frac_snow_veg could be made PFT dependent. 
         frac_snow_veg_loc(ipts,ivm) = frac_snow_veg(ipts)

         ! for this grid square and this vegetation type, we have a set LAI 
         ! effective
         laieff_collim_pft(1:nlevels_tot) = &
              laieff_collim(1:nlevels_tot,ipts,ivm)
         laieff_isotrop_pft(1:nlevels_tot) = &
              laieff_isotrop(1:nlevels_tot,ipts,ivm)

         IF(veget_max(ipts,ivm) == zero)THEN
            ! this vegetation type is not present, so no reason to do the 
            ! calculation
            CYCLE
         ENDIF

         DO ks = 1, n_spectralbands ! Loop over # of spectra

            ! set single scattering albedo and preferred scattering direction
            leaf_single_scattering_albedo = leaf_ssa(ivm,ks)
            leaf_psd_temp = leaf_psd(ivm,ks)                 

            ! calculate the rleaf and tleaf from wl=rl+tl and dl=rl/tl
            leaf_transmittance = leaf_single_scattering_albedo/ & 
                 (leaf_psd_temp+1)
            leaf_reflectance = leaf_psd_temp * &
                 leaf_transmittance

            ! We need to take into account the effect of snow on the 
            ! background reflectance here. From the snow above I can 
            ! calculate the true background reflectance for this PFT.
            ! The flag alb_bg_modis is a bit confusing because in 
            ! both cases the data source is the MODIS albedo product
            IF (alb_bg_modis) THEN

               ! If the flag is set to TRUE the model will read a 
               ! spatially explicit map of the background albedo for
               ! each PIXEL. The background albedo of the pixel is 
               ! then used to calculate the background albedo of the
               ! PFT accounting for snow fraction and snow albedo. Note
               ! that both snow fraction and snow albedo vary over time.
               br_base_temp = (un-frac_snow_veg_loc(ipts,ivm)) * & 
                 bckgrnd_alb(ipts,ks) + &
                 ( frac_snow_veg_loc(ipts,ivm)  ) * snowa_veg(ipts,ivm,ks)

            ELSE
               
               ! If the flag is set to FALSE the model uses the original
               ! parameterization at the PFT level. Each PFT has a fixed
               ! background albedo. The background albedo is recalculated
               ! taking into account the snowfraction and the snow albedo.
               ! Note that snow fraction and snoe albedo both vary over
               ! time.
               br_base_temp= (un-frac_snow_veg_loc(ipts,ivm)) * & 
                 bgd_reflectance(ivm,ks) + &
                 ( frac_snow_veg_loc(ipts,ivm)  ) * snowa_veg(ipts,ivm,ks)

            ENDIF

            ! At this step, we assume that there is no difference between the 
            ! direct and the diffuse background reflectance, which is true for 
            ! the background but not the canopy albedo
            br_base_temp_collim=br_base_temp
            br_base_temp_isotrop=br_base_temp

            ! Debug - Set flag for extra output
            IF(printlev_loc>=4 .AND. test_pft == ivm .AND. &
                 test_grid == ipts)THEN
               lprint=.TRUE.
            ELSE
               lprint=.FALSE.
            ENDIF
            !-
    
            ! Now solve the multilevel scheme 
            CALL multilevel_albedo(cosine_sun_angle, &
                 leaf_single_scattering_albedo, &
                 leaf_psd_temp, br_base_temp_collim, br_base_temp_isotrop, &
                 laieff_collim_pft, laieff_isotrop_pft, lconverged, &
                 Collim_Alb_Tot, Collim_Tran_Coll, Collim_Abs_Tot, &
                 Isotrop_Alb_Tot, &
                 Isotrop_Tran_Coll, Isotrop_Abs_Tot, Collim_Tran_Uncoll, &
                 Isotrop_Tran_Uncoll, lprint)

            Collim_Abs_Tot_mean(:) = Collim_Abs_Tot_mean(:) + Collim_Abs_Tot(:)
            Collim_Alb_Tot_mean(:) = Collim_Alb_Tot_mean(:) + Collim_Alb_Tot(:)

            ! This is a lot of information printed out for the paper
            ! on the multilevel albedo scheme (McGrath et al 2016, GMDD).  
            ! Probably not useful for anyone else, but I'm keeping it in 
            ! until the paper is published. This is why I'm protecting it 
            ! with an IF statement.  I do not use printlev_loc>=4 since it 
            ! really is too much information for normal usage.
            IF(.FALSE.)THEN

               WRITE(6,'(A,2F14.10)') 'Background reflectance ',&
                    br_base_temp_collim,br_base_temp_isotrop
               WRITE(6,'(A,F14.10)') 'Single scattering albedo (wl): ',&
                    leaf_single_scattering_albedo
               WRITE(6,'(A,F14.10)') 'Preferred scattering direction (dl): ',&
                    leaf_psd_temp

               WRITE(6,*) 'Collimated light ',ACOS(cosine_sun_angle)/pi*180.0
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        1  &  ',laieff_collim_1,&
                    ' & ',Collim_Alb_Tot_1,' & ',&
                    Collim_Tran_Tot_1-Collim_Tran_Uncollided_1,' & ',&
                    Collim_Tran_Uncollided_1,' & ',&
                    Collim_Alb_Tot_1+Collim_Tran_Tot_1,' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        2  &  ',laieff_collim_1,&
                    ' & ',Collim_Alb_Tot(nlevels_tot),' & ',&
                    Collim_Tran_Coll(1),' & ',Collim_Tran_Uncoll(1),' & ',&
                    Collim_Alb_Tot(nlevels_tot)+Collim_Tran_Coll(1)+&
                    Collim_Tran_Uncoll(1),' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '     Diff  &  ',laieff_collim_1,&
                    ' & ',ABS(Collim_Alb_Tot_1-Collim_Alb_Tot(nlevels_tot)),' & ',&
                    ABS((Collim_Tran_Tot_1-Collim_Tran_Uncollided_1)-&
                    Collim_Tran_Coll(1)),' & ',&
                    ABS(Collim_Tran_Uncollided_1-Collim_Tran_UnColl(1)),' & ',&
                    ABS((Collim_Alb_Tot_1+Collim_Tran_Tot_1)-&
                    (Collim_Alb_Tot(nlevels_tot)+Collim_Tran_Coll(1)+&
                    Collim_Tran_Uncoll(1))),' \\ '

               WRITE(6,*) 'Isotropic light '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        1  &  ',&
                    laieff_isotrop_1,' & ',Isotrop_Alb_Tot_1,' & ',&
                    Isotrop_Tran_Tot_1-Isotrop_Tran_Uncollided_1,' & ',&
                    Isotrop_Tran_Uncollided_1,' & ',&
                    Isotrop_Alb_Tot_1+Isotrop_Tran_Tot_1,' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '        2  &  ',&
                    laieff_isotrop_1,' & ',Isotrop_Alb_Tot(nlevels_tot),' & ',&
                    Isotrop_Tran_Coll(1),' & ',Isotrop_Tran_Uncoll(1),' & ',&
                    Isotrop_Alb_Tot(nlevels_tot)+Isotrop_Tran_Coll(1)+&
                    Isotrop_Tran_Uncoll(1),' \\ '
               WRITE(6,'(A,F4.1,A,4(F10.6,A))') '     Diff  &  ',&
                    laieff_isotrop_1,&
                    ' & ',ABS(Isotrop_Alb_Tot_1-Isotrop_Alb_Tot(nlevels_tot)),' & ',&
                    ABS((Isotrop_Tran_Tot_1-Isotrop_Tran_Uncollided_1)-&
                    Isotrop_Tran_Coll(1)),' & ',&
                    ABS(Isotrop_Tran_Uncollided_1-Isotrop_Tran_UnColl(1)),' & ',&
                    ABS((Isotrop_Alb_Tot_1+Isotrop_Tran_Tot_1)-&
                    (Isotrop_Alb_Tot(nlevels_tot)+Isotrop_Tran_Coll(1)+&
                    Isotrop_Tran_Uncoll(1))),' \\ '

               WRITE(6,1010) 'COLLIMTRAN ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_tot),1,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Collim_Tran_Tot_1,(Collim_Tran_Coll(1)+Collim_Tran_Uncoll(1))
               WRITE(6,1010) 'COLLIMALB ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_tot),2,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Collim_Alb_Tot_1,Collim_Alb_Tot(nlevels_tot)
               WRITE(6,1010) 'COLLIMABSCAN ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_tot),3,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Collim_Abs_Tot_1,SUM(Collim_Abs_Tot(:))
               WRITE(6,1010) 'COLLIMABSSOIL ',laieff_collim_1, &
                    laieff_collim_pft(nlevels_tot),4,&
                    leaf_single_scattering_albedo,br_base_temp_collim,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    un-Collim_Alb_Tot_1-Collim_Abs_Tot_1,un-&
                    Collim_Alb_Tot(nlevels_tot)-SUM(Collim_Abs_Tot(:))

               WRITE(6,1010) 'ISOTROPTRAN ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_tot),1,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Isotrop_Tran_Tot_1,(Isotrop_Tran_Coll(1)+Isotrop_Tran_Uncoll(1))
               WRITE(6,1010) 'ISOTROPALB ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_tot),2,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Isotrop_Alb_Tot_1,Isotrop_Alb_Tot(nlevels_tot)
               WRITE(6,1010) 'ISOTROPABSCAN ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_tot),3,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    Isotrop_Abs_Tot_1,SUM(Isotrop_Abs_Tot(:))
               WRITE(6,1010) 'ISOTROPABSSOIL ',laieff_isotrop_1, &
                    laieff_isotrop_pft(nlevels_tot),4,&
                    leaf_single_scattering_albedo,br_base_temp_isotrop,&
                    leaf_psd_temp,ACOS(cosine_sun_angle)/pi*180.0,&
                    un-Isotrop_Alb_Tot_1-Isotrop_Abs_Tot_1,un-&
                    Isotrop_Alb_Tot(nlevels_tot)-SUM(Isotrop_Abs_Tot(:))

1010           FORMAT(A,2(F12.6,2X),I4,2X,6(F14.8,2X))

            ENDIF ! debugging

            ! For our total albedo, we are just taking a simple mean of the 
            ! diffuse and the direct light, which means we lose some 
            ! information. This might be changed in the future. We also need 
            ! to weight this by the percentage of nonbio land, but this 
            ! appears to be included in veget_max
            converged_albedo = (Collim_Alb_Tot(nlevels_tot) + &
                 Isotrop_Alb_Tot(nlevels_tot))/2.0_r_std
            albedo(ipts,ks) = albedo(ipts,ks) + &
                 veget_max(ipts,ivm)*converged_albedo
            albedo_pft(ipts,ivm,ks)=veget_max(ipts,ivm)*converged_albedo

            ! Save the absorbed radiation for photosynthesis, we need only the
            ! visible range and use the isotropic part, this can be change 
            ! later to distinguish between sunlit and shaded leaves.
            ! Be careful here
            IF (ks == 1) THEN
               ! Test if Collim_Abs_Tot has negative values
               ! If so, set it to min_sechiba  
               DO  ilevel=1,nlevels_tot
                  IF (Isotrop_Abs_Tot(ilevel) .LT. zero) THEN
                     Isotrop_Abs_Tot(ilevel) =  min_sechiba
                  ENDIF
               ENDDO
               !
               Isotrop_Abs_Tot_p(ipts,ivm,:) = Isotrop_Abs_Tot(:)
               Isotrop_Tran_Tot_p(ipts,ivm,:) = Isotrop_Tran_Coll(:) + &
                    Isotrop_Tran_Uncoll(:)

               ! Notice that this light is actually cumulative, not per
               ! level!  This was needed for debugging purposes and running
               ! tests.  However, the photosynthesis routines need the light
               ! transmitted per level, i.e. if there is zero LAI
               ! in a level, the transmitted light will be one.
               DO ilevel=1,nlevels_tot-1
                  IF(Isotrop_Tran_Tot_p(ipts,ivm,ilevel+1) .GT. alb_threshold)THEN
                     Isotrop_Tran_Tot_p(ipts,ivm,ilevel)=&
                          Isotrop_Tran_Tot_p(ipts,ivm,ilevel) / &
                          Isotrop_Tran_Tot_p(ipts,ivm,ilevel+1)
                  ELSE

                     ! Here, we really don't know anything about how much 
                     ! light is transmitted in this layer, but there is no 
                     ! light reaching it from above so we can safely assume 
                     ! no photosynthesis takes place.  This is equivalent
                     ! to assuming it has no LAI, which means the 
                     ! transmission will be unity.
                     Isotrop_Tran_Tot_p(ipts,ivm,ilevel)=un

                  ENDIF
               ENDDO

               ! Debug
               IF (printlev_loc>=4 .AND. ivm == test_pft .AND. &
                    ipts == test_grid) THEN
                  WRITE(numout,*) 'Absorbed light in albedo.f90: ', ipts, ivm
                  DO ilevel=nlevels_tot,1,-1
                     WRITE(numout,'(I5,10F20.10)') ilevel, &
                          Isotrop_Abs_Tot_p(ipts,ivm,ilevel),&
                          Isotrop_Tran_Tot_p(ipts,ivm,ilevel),&
                          laieff_isotrop(ilevel,ipts,ivm),&
                          laieff_collim(ilevel,ipts,ivm)
                  ENDDO
               ENDIF
               !-

            ENDIF ! IF ks==1

         ENDDO ! ks = 1, n_spectralbands 

      ENDDO ! ivm = 2, nvm  

   ENDDO ! ipts = 1, npts 

   ! now we need to average the albedo over all our spectra, so we can 
   ! pass it to modules which do not distinguish between different 
   ! spectral bands
   albedo_glob(:) = zero
   DO ks = 1, n_spectralbands ! Loop over # of spectra
      albedo_glob(:) = albedo_glob(:) + albedo(:,ks)
   ENDDO
   albedo_glob(:)=albedo_glob(:)/REAL(n_spectralbands,r_std)

   Collim_Abs_Tot_mean(:) = Collim_Abs_Tot_mean(:) / n_spectralbands
   Collim_Alb_Tot_mean(:) = Collim_Abs_Tot_mean(:) / n_spectralbands

   ! Error checking
   IF (abs(albedo(1,1)) .gt. 1.0d0) THEN

       CALL print_debugging_albedo_info(1, cosine_sun_angle, &
             leaf_reflectance, leaf_transmittance, laieff_collim_1, &
             laieff_isotrop_1, br_base_temp_collim, br_base_temp_isotrop, &
             Collim_Alb_Tot_1, Collim_Tran_Tot_1, Collim_Abs_Tot_1, &
             Isotrop_Alb_Tot_1, Isotrop_Tran_Tot_1, Isotrop_Abs_Tot_1)
       STOP

   END IF
   !-

   ! Debug
   IF (printlev_loc>=4) THEN
      DO ivm = 1,nvm
         WRITE(numout,*) 'laieff_isotrop end albedo for ivm',ivm,'is', &
              laieff_isotrop(:,:,ivm)
      ENDDO
   ENDIF
   !-

   IF (printlev>=3) WRITE(numout,*) 'Leaving albedo_two_stream'

 END SUBROUTINE albedo_two_stream

!! ==============================================================================================================================
!! SUBROUTINE   : twostream_solver
!!
!>\BRIEF        : Computes the two-stream albedo solution for a level with given single scatterer properties
!!                and a defined background albedo
!!
!! DESCRIPTION  : This solution is the two-stream solution of Pinty et al (2006) for a vegetation level above
!!                an isotropically reflecting background.  It breaks down the problem into three parts, solving
!!                each part for the case of diffuse (isotropic) and direct (collimated) light.  The three parts are,
!!                1) The term due to light that does not interact at all with the canopy (Black Canopy)
!!                2) Light that does not interact at all with the background (Black Background)
!!                3) Light that bounces between the background and the canopy
!!                This routine (and the routines it uses) were received directly from Bernard Pinty, and only some
!!                minor modifcations were made, in addition to more documentation
!! 
!!                NOTE: the single layer solution is no longer used. The code is kept here in case it is decided that
!!                      the multi layer solution is two expensive if only a single layer is used for the energy budget.
!!                      In that case a solution needs to be implemeted to calculate the light distribution within
!!                      the canopy. 
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S): Collim_Alb_Tot,Collim_Tran_Tot,
!!      Collim_Abs_Tot,Isotrop_Alb_Tot,Isotrop_Tran_Tot,Isotrop_Abs_To
!!
!! REFERENCE(S) :  B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron and M. M. Verstraete (2006).
!! Simplifying the Interaction of Land Surfaces with Radiation for Relating Remote Sensing Products to Climate Models.
!! Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================

 SUBROUTINE twostream_solver(leaf_reflectance, leaf_transmittance, background_reflectance_collim, background_reflectance_isotrop, &
          cosine_sun_angle, Collim_Alb_Tot, Collim_Tran_Tot, Collim_Abs_Tot, &
          Isotrop_Alb_Tot, Isotrop_Tran_Tot, Isotrop_Abs_Tot, laieff_collim, &
           laieff_isotrop,  Collim_Tran_Uncollided,  Isotrop_Tran_Uncollided)


   !! 0. Variables and parameter declaration
   !! 0.1 Input variables
   REAL(r_std), INTENT(IN)    :: leaf_reflectance            !! effective leaf reflectance, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: leaf_transmittance          !! effective leaf transmittance, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: background_reflectance_collim      !! background reflectance for direct radiation,
                                                             !! between 0 and 1
   REAL(r_std), INTENT(IN)    :: background_reflectance_isotrop      !! background reflectance for diffuse radiation,
                                                             !! between 0 and 1
   REAL(r_std), INTENT(IN)    :: cosine_sun_angle            !! cosine of the solar zenith angle, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: laieff_collim               !! Effective Leaf Area Index, computed at current sun angle
   REAL(r_std), INTENT(IN)    :: laieff_isotrop              !! Effective Leaf Area Index, computed at 60 degrees
                                                             !! @tex $(m^{2} m^{-2})$ @endtex 

   !! 0.2 Output variables
   !! notice that these variables (absorption + transmission + reflection) do not necessarily add up to 1 due to multiple scattering
   REAL(r_std), INTENT(OUT)   :: Collim_Alb_Tot              !! collimated total albedo from this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Collim_Tran_Tot             !! collimated total transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Collim_Abs_Tot              !! collimated total absorption by this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Alb_Tot             !! isotropic total albedo from this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Tran_Tot            !! isotropic total transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Abs_Tot             !! isotropic total absorption by this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Collim_Tran_Uncollided      !! collimated uncollied transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(OUT)   :: Isotrop_Tran_Uncollided     !! isotropic uncollied transmission through this level, between 0 and 1
                                                             !! @tex $()$ @endtex 

   !! 0.3 Modified variables

   !! 0.4 Local variables
   LOGICAL :: ok
   REAL(r_std), DIMENSION(4)                           :: gammaCoeffs
   REAL(r_std), DIMENSION(4)                           :: gammaCoeffs_star
   REAL(r_std)                                         :: tauprimetilde
   REAL(r_std)                                         :: tauprimestar
   REAL(r_std)                                         :: sun_zenith_angle_radians
   REAL(r_std), PARAMETER                              :: isotropic_cosine_constant=0.5/0.705

   !calculated fluxes

   REAL(r_std)                :: Collim_Alb_BB 
   REAL(r_std)                :: Collim_Tran_BB
   REAL(r_std)                :: Collim_Abs_BB
   REAL(r_std)                :: Isotrop_Alb_BB
   REAL(r_std)                :: Isotrop_Tran_BB
   REAL(r_std)                :: Isotrop_Abs_BB
   REAL(r_std)                :: Collim_Tran_BC
   REAL(r_std)                :: Isotrop_Tran_BC
   REAL(r_std)                :: Collim_Tran_TotalOneWay
   REAL(r_std)                :: Isotrop_Tran_TotalOneWay
   REAL(r_std)                :: Below_reinject_rad_collim,Below_reinject_rad_isotrop

!_ ================================================================================================================================

   IF (printlev>=2) WRITE(numout,*) 'Entering twostream_solver'

   ! convert angular values
   
   sun_zenith_angle_radians = acos(cosine_sun_angle)

   ! calculate the 4 gamma coefficients both in isotropic and collimated illumination 

   call gammas(leaf_reflectance, leaf_transmittance, cosine_sun_angle, gammaCoeffs)
   call gammas(leaf_reflectance,leaf_transmittance,isotropic_cosine_constant,&
        gammaCoeffs_star)

   ! estimate the effective value of the optical thickness 

   tauprimetilde = 0.5_r_std * laieff_collim
   !tauprimetilde = 1.

   ! Should become zenit angle dependent (=60°) calculated by the Monte Carlo 
   ! photon shooting routine

   tauprimestar  = 0.5_r_std * laieff_isotrop
   !tauprimestar  = 1.

   ! +++++++++++++ BLACK BACKGROUND ++++++++++++++++++++++ 
   ! * Apply the black-background 2stream solution 
   ! * These equations are written for the part of the incoming radiation 
   ! * that never hits the background but does interact with the vegetation 
   ! * canopy.
   ! * 
   ! * Note : the same routine dhrT1() is used for both the isotropic and
   ! * collimated illumination conditions but the calling arguments differ.
   ! * (especially the solar angle used).
   ! */

   !	/* 1) collimated source */
   ok=dhrT1(leaf_reflectance,leaf_transmittance,&
        gammaCoeffs(1),gammaCoeffs(2),gammaCoeffs(3),gammaCoeffs(4),&
        sun_zenith_angle_radians, tauprimetilde,&
        Collim_Alb_BB,Collim_Tran_BB,Collim_Abs_BB)                      
   ! 	/* 2) isotropic source */
   ok=dhrT1(leaf_reflectance,leaf_transmittance,&
        gammaCoeffs_star(1),gammaCoeffs_star(2),gammaCoeffs_star(3),&
        gammaCoeffs_star(4),&
        acos(isotropic_cosine_constant),tauprimestar,&
        Isotrop_Alb_BB,Isotrop_Tran_BB,Isotrop_Abs_BB)


   ! +++++++++++++ BLACK CANOPY ++++++++++++++++++++++ 
   ! * Apply the black canopy solution.
   ! * These equations hold for the part of the incoming radiation 
   ! * that do not interact with the vegetation, travelling through 
   ! * its gaps.
   ! */

   ! 	/* 1) collimated source */
   IF (cosine_sun_angle .NE. 0) THEN
      Collim_Tran_BC  = exp( - tauprimetilde/cosine_sun_angle)
      Collim_Tran_Uncollided = Collim_Tran_BC
   ENDIF


   ! 	/* 2) isotropic source */
   Isotrop_Tran_BC = TBarreUncoll_exact(tauprimestar)
   Isotrop_Tran_Uncollided = Isotrop_Tran_BC


   ! /* Total one-way transmissions:
   ! * The vegetation canopy is crossed (one way) by the uncollided radiation
   ! * (black canopy) and the collided one (black background). */

   !	/* 1) collimated source */
   Collim_Tran_TotalOneWay  = Collim_Tran_BC  + Collim_Tran_BB 

   !	/* 2) isotropic source */
   Isotrop_Tran_TotalOneWay = Isotrop_Tran_BC + Isotrop_Tran_BB 

   ! * The Below_reinject_rad describes the process of reflecting toward the 
   ! * background the upward travelling radiation (re-emitted from below the 
   ! * canopy). It appears in the coupling equations as the limit of the series: 
   ! *    1 + rg*rbv + (rg*rbv)^2 + (rg*rbv)^3 + ...
   ! *      where rg is the background_reflectance and rbv is Isotrop_Alb_BB
   ! *      (with Isotrop describing the Lambertian reflectance of the background).
   ! */
   ! this might be involved in Eq. 27, 
   Below_reinject_rad_collim = un / (un - background_reflectance_collim*Isotrop_Alb_BB)
   Below_reinject_rad_isotrop = un / (un - background_reflectance_isotrop*Isotrop_Alb_BB)

   !/* TOTAL ALBEDO */
   ! 	/* 1) collimated source */
   Collim_Alb_Tot = Collim_Alb_BB + &
        background_reflectance_collim * Collim_Tran_TotalOneWay * &
        Isotrop_Tran_TotalOneWay * Below_reinject_rad_collim
   !	/* 2) isotropic source */
   Isotrop_Alb_Tot = Isotrop_Alb_BB + & 
        background_reflectance_isotrop * Isotrop_Tran_TotalOneWay * &
        Isotrop_Tran_TotalOneWay * Below_reinject_rad_isotrop ! this seems to be
   ! exactly Eq. 33


   !/* TOTAL TRANSMITION TO THE BACKGROUND LEVEL */
   !	/* 1) collimated source */
   Collim_Tran_Tot = Collim_Tran_TotalOneWay * Below_reinject_rad_collim ;

   !	/* 2) isotropic source */
   Isotrop_Tran_Tot = Isotrop_Tran_TotalOneWay * Below_reinject_rad_isotrop ;

   !/* TOTAL ABSORPTION BY THE VEGETATION LEVEL */
   !	/* 1) collimated source */
   Collim_Abs_Tot = un - (Collim_Tran_Tot + Collim_Alb_Tot) + &
        background_reflectance_collim * Collim_Tran_Tot;
   !	/* 2) isotropic source */
   Isotrop_Abs_Tot = un - (Isotrop_Tran_Tot + Isotrop_Alb_Tot) + &
        background_reflectance_isotrop * Isotrop_Tran_Tot;

!+++ TEMP +++ 
! Combine the collimated & isotrop together as Collimate for energybudget 
   Collim_Alb_Tot = Collim_Alb_Tot + Isotrop_Alb_Tot
   Collim_Abs_Tot = Collim_Abs_Tot + Isotrop_Abs_Tot   
!+++ TEMP +++       
        

    IF (printlev>=3) WRITE(numout,*) 'Exiting twostream_solver'

 END SUBROUTINE twostream_solver


!! ============================================================================\n
!! FUNCTION     : dhrT1
!!
!>\BREF         
!!
!! DESCRIPTION : Taken directly from the Fortan code provided by Pinty et al for their \n
!!               two-stream model.  The only changes are the REAL types.  This is used
!!               to compute the black background absorption, transmission, and reflectance
!!               in Pintys scheme and it's based on the older model of Meador and Weaver...
!!               Pinty 2006 also includes a short discussion of it in Appendix A 
!!
!! RECENT CHANGE(S): None\n
!!
!! RETURN VALUE : dhrT1
!!
!! REFERENCE(S) : Meador and Weaver, 'Two-stream approximations to radiative transfer in
!!    planetary atmosphere: a unified description of existing methods and a new improvement'
!!    J. Atmospheric Sciences, VOL 37, p. 630--643, 1980.
!!
!! FLOWCHART    : None
!_ =============================================================================

FUNCTION dhrT1(rl,tl,gamma1,gamma2,gamma3,gamma4,tta,tau,AlbBS,Tdif,AbsVgt)


  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  REAL(r_std), INTENT(IN) :: rl  ! the effective reflectance of a single scatterer, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: tl  ! the effective transmittance of a single scatterer, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: gamma1 ! the gamma coefficients from Meador and Weaver
  REAL(r_std), INTENT(IN) :: gamma2
  REAL(r_std), INTENT(IN) :: gamma3
  REAL(r_std), INTENT(IN) :: gamma4
  REAL(r_std), INTENT(IN) :: tta ! the solar zenith angle, between 0 and pi/2
                                 !! @tex $(radians)$ @endtex      

  REAL(r_std), INTENT(IN) :: tau ! Effective LAI * G(theta)
                                 !! @tex $()$ @endtex      

  !! 0.2 Output variables
  REAL(r_std), INTENT(OUT) :: AlbBS ! the albedo of level, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(OUT) :: Tdif ! the transmitted light (all diffuse) from this light source, between 0 and 1
                                 !! @tex $()$ @endtex      
  REAL(r_std), INTENT(OUT) :: AbsVgt ! the light absorbed by the level, between 0 and 1
                                 !! @tex $()$ @endtex      
  LOGICAL :: dhrT1 ! the output flag, which always appears to be true

  !! 0.3 Modified variables
  !! 0.4 Local variables

  REAL(r_std) :: alpha1,alpha2,ksquare,k
  REAL(r_std) :: first_term,secnd_term1,secnd_term2,secnd_term3
  REAL(r_std) :: expktau,Tdir
  REAL(r_std) :: mu0,w0 

  !_ ================================================================================================================================

  mu0=cos(tta)
  w0=rl+tl


  Tdir = exp(-tau/mu0) ! direct transmission

  ! There is a difference between conservative and non-conservative scattering 
  ! conditions */
  IF (w0 .ne. 1.0 .AND. w0 .ne. 0.0) THEN
     !NON_CONSERVATIVE SCATTERING

     ! some additional parameters, taken from Eq. 16--18 of Meador and Weaver, J. Atmospheric Sciences,
     ! Vol 37, p. 630--643 (1980)

     alpha1  = gamma1*gamma4 + gamma2*gamma3
     alpha2  = gamma1*gamma3 + gamma2*gamma4
     ksquare = gamma1*gamma1 - gamma2*gamma2
     k       = sqrt(ksquare)

     expktau = exp(k*tau)

     !Black Soil Albedo...Eq. 14 in Meador and Weaver
     first_term  = ((un-ksquare*mu0*mu0)*((k+gamma1)*expktau + (k-gamma1)/expktau))
     IF (first_term .eq. 0.0) THEN
        !we will be dividing by zero : cannot continue.
        dhrT1 = .false.
     ELSE 
        first_term = un/first_term
        secnd_term1 = (un - k*mu0)*(alpha2 + k*gamma3)*expktau
        secnd_term2 = (un + k*mu0)*(alpha2 - k*gamma3)/expktau
        secnd_term3 = 2.0_r_std * k * (gamma3 - alpha2*mu0)*Tdir

        AlbBS = (w0 * first_term * (secnd_term1 - secnd_term2 - secnd_term3))

        !Transmission...Eq. 15 in Meador and Weaver, for diffuse light?
        IF (ksquare .eq. 0.0) THEN 
           first_term = un
        ENDIF
        secnd_term1 = (un+k*mu0)*(alpha1+k*gamma4)*expktau
        secnd_term2 = (un-k*mu0)*(alpha1-k*gamma4)/expktau
        secnd_term3 = 2.0_r_std * k * (gamma4 + alpha1*mu0)
        Tdif = - w0*first_term*(Tdir*(secnd_term1 - secnd_term2) - secnd_term3)

        ! Absorption by vegetation...whatever is not transmitted or reflected 
        ! must be absorbed
        AbsVgt = (un- (Tdif+Tdir) - AlbBS)
     ENDIF ! first_term .eq. 0.0
  ELSE IF (w0 .eq. 0.) THEN
     !BLACK CANOPY
     AlbBS = zero
     Tdif  = zero
     AbsVgt = un - Tdir
  ELSE
     !CONSERATIVE SCATTERING...Eq. 24 in Meador and Weaver
     AlbBS =  (un/(un + gamma1*tau))*(gamma1*tau + (gamma3-gamma1*mu0)*&
          (un-exp(-tau/mu0)));
     Tdif   = un - AlbBS - Tdir;
     AbsVgt = zero;
  ENDIF ! w0 .ne. 1.0 .AND. w0 .ne. 0.0

! not sure what the purpose of this flag is, as it will always be set to true here
  dhrT1 = .true.

END FUNCTION dhrT1

!! ============================================================================\n
!! FUNCTION     : TBarreUncoll_exact
!!
!>\BRIEF          Computes the transmission of the diffuse black canopy light
!!
!! DESCRIPTION : Taken directly from the Fortan code provided by Pinty et al for their \n
!!               two-stream model.  The only changes are the REAL types.  This appears
!!               to be solving Eq. 16 in Pinty 2006, which is the transmission which
!!               does not collide with the canopy
!!
!! RECENT CHANGE(S): None\n
!!
!! RETURN VALUE : TBarreUncoll_exact
!!
!! REFERENCE(S) : None
!!
!! FLOWCHART    : None
!_ =============================================================================

FUNCTION TBarreUncoll_exact(tau)


  !! 0. Variables and parameter declaration

  !! 0.1 Input variables
  REAL(r_std), INTENT(IN) :: tau ! Effective LAI * G(theta)
  !! @tex $()$ @endtex      

  !! 0.2 Output variables
  REAL(r_std) :: TBarreUncoll_exact ! the isotropic light transmission uncollided
                                    ! with the canopy, between 0 and 1
  !! @tex $()$ @endtex      


  !! 0.3 Modified variables

  !! 0.4 Local variables
!  INTEGER :: j,ind
!  REAL(r_std) :: iGammaloc
!  INTEGER :: order

!_ ================================================================================================================================

  !+++++ CHECK +++++

!!$  iGammaloc = zero
!!$  order=20
!!$
!!$  ! in the case where tau is equal to zero, this crashes in the first loop where
!!$  ! iGammaloc is also zero...quick fix, and might even be accurate
!!$
!!$  IF(tau .GT. 1e-10_r_std)THEN
!!$
!!$     DO j=0,order-1
!!$        ind = order - j
!!$        iGammaloc = ind / (un + ind/(tau+iGammaloc))
!!$     END DO
!!$     iGammaloc=un/(tau + iGammaloc)
!!$  ENDIF
!!$
!!$  TBarreUncoll_exact = exp(-tau)*(un - tau + tau*tau*iGammaloc)

  ! this is a change suggested by Bernard to improve the matching between the one
  ! level and multilevel case

  TBarreUncoll_exact = exp(-tau*2.0_r_std*0.705_r_std)
  !++++++++++

END FUNCTION TBarreUncoll_exact

!! ==============================================================================================================================
!! SUBROUTINE   : gammas
!!
!>\BRIEF         Computes a set of gamma coefficients for use in the black background equations
!!
!! DESCRIPTION : Taken directly from the Fortan code provided by Pinty et al for their \n
!!               two-stream model.  The only changes are the REAL types.  This calculates
!!               the gamma coefficients based on Appendix A in the paper.  This seems to
!!               make the assumption of the Ross's G function and a spherical leaf
!!               angle distribution.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): gammaCoeff
!!
!! REFERENCE(S) :  B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron and M. M. Verstraete (2006).
!! Simplifying the Interaction of Land Surfaces with Radiation for Relating Remote Sensing Products to Climate Models.
!! Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!_ ================================================================================================================================

SUBROUTINE gammas(rl, tl, mu0, gammaCoeff)


  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  REAL(r_std), INTENT(IN) :: rl  ! the effective reflectance of a single scatterer, between 0 and 1
  !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: tl  ! the effective transmittance of a single scatterer, between 0 and 1
  !! @tex $()$ @endtex      
  REAL(r_std), INTENT(IN) :: mu0 ! the cosine of the solar zenith angle, between 0 and 1
  !! @tex $()$ @endtex      

  !! 0.2 Output variables
  REAL(r_std), DIMENSION(1:4), INTENT(OUT) :: gammaCoeff ! the set of gamma coefficients in the above reference
  !! @tex $()$ @endtex      

  !! 0.3 Modified variables

  !! 0.4 Local variables
  REAL(r_std) :: wd,w0,w0half,wdsixth

!_ ================================================================================================================================


  w0      = rl+tl
  wd      = rl-tl
  w0half  = w0*0.5_r_std
  wdsixth = wd/6.0_r_std

  gammaCoeff(1)=2._r_std*(un - w0half + wdsixth)
  gammaCoeff(2)=2._r_std*(w0half + wdsixth)
  gammaCoeff(3)=2._r_std*(w0half*0.5_r_std + mu0*wdsixth)/w0
  gammaCoeff(4)=un-gammaCoeff(3)

END SUBROUTINE gammas


!! ==============================================================================================================================
!! SUBROUTINE   : calculate_snow_albedo
!!
!>\BRIEF         Computes some of the information needed to calculate the effect of the snow albedo
!!               on the background reflectance in the two stream model.  Right now, this can be done
!!               with the old snow albedo scheme from Krinner et al 2005 as well as the snow albedo
!!               scheme from Community Land Model 3 (CLM3). The calculation of
!!               snow cover fraction is taken from Yang et al. 1997
!!
!! DESCRIPTION : In order to compute the background albedo in Pinty's two stream model, we have to take
!!               into account any snow that has fallen through the canopy and landed on the ground.  In particular,
!!               we need the amount of ground covered by the snow and the albedo of this snow.  Both of these
!!               quantities are calculated here, using a function that changes the albedo of the snow based on
!!               its age.  This is done in two ways, but the model of CLM3 is better suited to be used with Pinty's
!!               albedo scheme, as it divides the radiation into VIS and NIR light. The calculation of snow cover 
!!               fraction depends on roughness lenght.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): ::snowa_veg, ::frac_snow_veg, ::albedo_snow
!!
!! REFERENCE(S) :  "Technical description of the community land model CLM)", K. W. Oleson, Y. Dai, G. Bonan, M. Bosilovich, et al.,
!!                 NCAR Technical Note, May 2004.
!!
!!                 Yang Z, Dickinson R, Robock A, Vinnikov K (1997) Validation of the snow submodel of the 
!!                 biosphere-atmosphere transfer scheme with Russian 
!!                 snow cover and meteorological observational data. Journal of Climate, 10, 353–373.
!!
!! FLOWCHART    : None
!_ ================================================================================================================================

SUBROUTINE calculate_snow_albedo(npts,  coszang,  snow,  snow_nobio, &
           snow_age,  snow_nobio_age,  frac_nobio,  albedo_snow, &
           snowa_veg,  frac_snow_veg,  snowa_nobio,  frac_snow_nobio, &
           veget_max, z0m)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
   INTEGER,INTENT(IN)                              :: npts            !! Domain size - Number of land pixels  (unitless) 
   REAL(r_std), DIMENSION(npts), INTENT(in)        :: coszang          !!  cosine of the solar zenith angle, between 0 and 1
                                                                      !! @tex $()$ @endtex 
   REAL(r_std),DIMENSION (npts), INTENT(in)        :: snow            !! Snow mass in vegetation (kg m^{-2})           
   REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: snow_nobio      !! Snow mass on continental ice, lakes, etc. (kg m^{-2})      
   REAL(r_std),DIMENSION (npts), INTENT(in)        :: snow_age        !! Snow age (days)        
   REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: snow_nobio_age  !! Snow age on continental ice, lakes, etc. (days)    
   REAL(r_std),DIMENSION (npts,nnobio), INTENT(in) :: frac_nobio      !! Fraction of non-vegetative surfaces, i.e. 
                                                                      !!    continental ice, lakes, etc. (unitless)     
   REAL(r_std), DIMENSION(npts,nvm), INTENT(in)    :: veget_max       !! PFT coverage fraction of a PFT (= ind*cn_ind) 
                                                                      !!    (m^2 m^{-2})
   REAL(r_std), DIMENSION(npts), INTENT(in)        :: z0m             !! Roughness height for momentum of vegetated part (m)
   REAL(r_std), DIMENSION(:),INTENT(in)            :: frac_snow_veg   !! The fraction of the surface for each PFT covered by snow
   REAL(r_std), DIMENSION(:,:),INTENT(in)          :: frac_snow_nobio !! The fraction of nonbiological land types covered by snow

  !! 0.2 Output variables

   REAL(r_std), DIMENSION(npts),INTENT(OUT)        :: albedo_snow     !! the averaged snow albedo for a given grid cell
   REAL(r_std), DIMENSION(npts,nvm,n_spectralbands),&
                                INTENT(OUT)        :: snowa_veg       !! the albedo of snow...this is seperated into
                                                                       !! PFT types, even though the calculation is identical
                                                                       !! for all PFTs right now
   REAL(r_std), DIMENSION(npts,nnobio,n_spectralbands),&
                                INTENT(OUT)        :: snowa_nobio     !! the albedo of snow on nonbiological land types


  !! 0.3 Modified variables

  !! 0.4 Local variables
   REAL(r_std) :: agefunc,agefunc_nobio,snow_alb_direct, snow_alb_diffuse,f_mu
   REAL(r_std),DIMENSION(n_spectralbands)            :: c_albedo,alb_snow_0 ! parameter values, taken directly from CLM3

   INTEGER :: ipts,ks,ivm,jv ! indices

   LOGICAL                                           :: do_new_snow_albedo = .TRUE.

  
  !_ ================================================================================================================================

  
  ! initialize the output
  albedo_snow=val_exp
  snowa_veg=val_exp
  snowa_nobio=val_exp


  ! set some values for the new snow albedo
  alb_snow_0(ivis)=0.95_r_std
  alb_snow_0(inir)=0.65_r_std
  c_albedo(ivis)=0.2_r_std
  c_albedo(inir)=0.5_r_std


  DO ipts = 1, npts ! loop over all the grid squares

     DO ivm=1,nvm ! loop over all the PFTs

        IF(veget_max(ipts,ivm) == zero)THEN
             ! this vegetation type is not present, so no reason to do the 
             ! calculation
             CYCLE
        ENDIF

        DO ks=1,n_spectralbands ! loop over spectra

            IF (ABS(fixed_snow_albedo - undef_sechiba) .GT. EPSILON(undef_sechiba)) THEN

               snowa_veg(ipts,ivm,ks) = fixed_snow_albedo

            ELSE

               IF(do_new_snow_albedo)THEN
                   ! first we calculate diffuse albedo
                   ! NOTE: The difference between these two methods has not been 
                   ! tested.  The only constraint is that it must be dimensionless
                   agefunc = un - EXP(-snow_age(ipts)/tcst_snowa) ! already in ORCHIDEE
                   !agefunc = zero
  

                   ! using the constant from CLM3 conveted into days**-1
                   !agefunc=un-un/(un+snow_age(ipts)*0.864_r_std)
                   !agefunc=un/(un+snow_age(ipts)*0.864_r_std)     
                   snow_alb_diffuse=(un-c_albedo(ks)*agefunc)*alb_snow_0(ks)

                   ! now the direct albedo
                   IF(coszang(ipts) .GT. 0.5_r_std)THEN
                       f_mu=zero
                   ELSE
                       ! substituting b=2.0 into the equation...recommended by CLM3
                       f_mu=(3.0_r_std)/(un+coszang(ipts))-2.0_r_std 
                   ENDIF

                   snow_alb_direct=snow_alb_diffuse+0.4_r_std*f_mu*&
                       (un-snow_alb_diffuse)

                   ! assume that the total snow albedo is a mix of 
                   ! 50% direct and 50% diffuse
                   snowa_veg(ipts,ivm,ks)=(snow_alb_direct+snow_alb_diffuse)/2.0_r_std

              ELSE
                   ! calculate the age of the snow for this grid square
                   agefunc = EXP(-snow_age(ipts)/tcst_snowa)
   
                   ! now the albedo...use only the bare soil snow decay parameters, 
                   ! since the true albedo is calculated
                   ! by the PFT-dependent two-stream solver
                   snowa_veg(ipts,ivm,ks) = snowa_aged(1)+snowa_dec(1)*agefunc 
     
              ENDIF ! control%do_new_snow_albedo

           ENDIF ! prescribe or calculate albedo

           ! Notice that the fraction of nobio land is included in veget_max.
           ! We have to average by the number of spectras, too, for this array, 
           ! because it doesn't distinguish between PFTs or spectra
           albedo_snow(ipts) = albedo_snow(ipts) + &
                frac_snow_veg(ipts) * veget_max(ipts,ivm) * &
                snowa_veg(ipts,ivm,ks)/REAL(n_spectralbands,r_std)

        ENDDO ! ks=1,n_spectralbands 

     ENDDO ! ivm=1,nvm 

     ! there is no distinction between NIR and VIS for non bio surfaces
     IF (ABS(fixed_snow_albedo - undef_sechiba) .GT. EPSILON(undef_sechiba)) THEN
        snowa_nobio(:,:,:) = fixed_snow_albedo
     ELSE

        DO jv = 1, nnobio 
           ! calculate the snow age
           agefunc_nobio = EXP(-snow_nobio_age(ipts,jv)/tcst_snowa)

           ! and the albedo due to snow of this age on this nonbio surface
           snowa_nobio(ipts,jv,:) = ( snowa_aged(1) + snowa_dec(1) * agefunc_nobio ) 
        ENDDO
     ENDIF ! prescribe or calculate

     DO jv = 1, nnobio 

        ! not sensitive to the spectral band.
        ! Don't need to average here since we are outside the loop over the
        ! bands
        albedo_snow(ipts) = albedo_snow(ipts) + &
             ( frac_nobio(ipts,jv) ) * ( frac_snow_nobio(ipts,jv) ) * &
             snowa_nobio(ipts,jv,1) 

     ENDDO ! jv = 1, nnobio 
     
  ENDDO ! ipts = 1, npts 


END SUBROUTINE CALCULATE_SNOW_ALBEDO

!! ==============================================================================================================================
!! SUBROUTINE   : optimize_albedo_values
!!
!>\BRIEF         Follow the radiation scattered through the canopy in the 
!!               case of multiple levels.
!!
!! DESCRIPTION : Right now, we know that using Pintys method in the case of a 
!!               single canopy level will result in different top of the canopy
!!               albedos than iwhat is found if we use multiple canopy levels.
!!               We trust that the single level case gives the 'true' values.
!!
!!               This algorithm follows light as it scatters through multiple
!!               levels in the canopy.  At each level, the scattering 
!!               solution is found by solving Pinty's two stream model.  This
!!               means that light which enters a level can either be transmitted
!!               without colliding with the vegetation, transmitted after
!!               collision with the vegetation, reflecting off the vegetation,
!!               or being absorbed.  We follow the fluxes until they get
!!               really small.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): ::lconverged, ::Collim_Alb_Tot, ::Collim_Tran_Tot, 
!!          ::Collim_Abs_Tot, ::Isotrop_Alb_Tot, ::Isotrop_Tran_Tot, ::Isotrop_Abs_Tot
!!
!! REFERENCE(S) :  B. Pinty, T. Lavergne, R.E. Dickinson, J-L. Widlowski, N. Gobron
!!          and M. M. Verstraete (2006). Simplifying the Interaction of Land 
!!          Surfaces with Radiation for Relating Remote Sensing Products to 
!!          Climate Models. Journal of Geophysical Research. Vol 111, D02116.
!!
!! FLOWCHART    : None
!_ ================================================================================================================================
SUBROUTINE multilevel_albedo(cosine_sun_angle, leaf_single_scattering_albedo_start,&
     leaf_psd_start, br_base_temp_collim,  br_base_temp_isotrop, &
     laieff_collim_pft, laieff_isotrop_pft, lconverged, &
     Collim_Alb_Coll, Collim_Tran_Coll, Collim_Abs_Tot, Isotrop_Alb_Coll, &
     Isotrop_Tran_Coll, Isotrop_Abs_Tot, Collim_Tran_Uncoll, Isotrop_Tran_Uncoll, lprint)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  REAL(r_std), INTENT(IN)    :: cosine_sun_angle                          !! cosine of the solar zenith angle, between 0 and 1
  !! @tex $()$ @endtex 
  REAL(r_std), INTENT(IN)    :: leaf_single_scattering_albedo_start       !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)    :: leaf_psd_start                            !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)    :: br_base_temp_collim                       !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)    :: br_base_temp_isotrop                      !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(IN)        :: laieff_collim_pft         !! Effective lai for a single pixel and pft to be 
  REAL(r_std),DIMENSION(:),INTENT(IN)        :: laieff_isotrop_pft        !! Effective lai for a single pixel and pft to be 
  LOGICAL,INTENT(IN)        :: lprint                                     !! A flag to print some debug statements
  !!  used in the two stream approach...every level

  !! 0.2 Output variables
  LOGICAL,INTENT(OUT) :: lconverged                                       !! did the optimization converge?
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Alb_Coll        !! collimated total albedo for the converged case
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Tran_Coll       !! collimated total transmission
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Abs_Tot         !! collimated total absorption
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Alb_Coll       !! isotropic total albedo
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Tran_Coll      !! isotropic total transmission
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Abs_Tot        !! isotropic total absorption
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Collim_Tran_Uncoll     !! collimated uncollided transmission
  !! unitless, between 0 and 1
  REAL(r_std),DIMENSION(:),INTENT(OUT)          :: Isotrop_Tran_Uncoll    !! isotropic uncollided transmission
  !! unitless, between 0 and 1

  !! 0.3 Modified variables

  !! 0.4 Local variables
  ! use an extra level here...this is basically to store the transmission from the sun, which is normalized to 1.0
  REAL(r_std),DIMENSION(nlevels_tot)            :: Collim_Abs_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Collim_Alb_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Collim_Tran_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Collim_Tran_UnColl_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Isotrop_Abs_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Isotrop_Alb_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Isotrop_Tran_Coll_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Isotrop_Tran_UnColl_Unscaled
  REAL(r_std),DIMENSION(nlevels_tot)            :: Collim_Tran_Tot
  REAL(r_std),DIMENSION(nlevels_tot)            :: Isotrop_Tran_Tot

  REAL(r_std),DIMENSION(0:nlevels_tot+1)            :: &
       collim_down_cs,isotrop_down_cs,isotrop_up_cs

  INTEGER                                       :: istep,ilevel

  REAL(r_std)                                   :: leaf_reflectance
  REAL(r_std)                                   :: leaf_transmittance

  LOGICAL :: lexit
  LOGICAL :: ok
  REAL(r_std), DIMENSION(4)                     :: gammaCoeffs
  REAL(r_std), DIMENSION(4)                     :: gammaCoeffs_star
  REAL(r_std)                                   :: sun_zenith_angle_radians
  REAL(r_std), PARAMETER                        :: isotropic_cosine_constant=0.5/0.705

  INTEGER,SAVE :: icalls=0
  !_ ================================================================================================================================


  icalls=icalls+1

  ! convet the sun angle
  sun_zenith_angle_radians = acos(cosine_sun_angle)



  ! initialize some values that are used in the optimization loop
  istep=0
  lexit=.FALSE.
  lconverged=.FALSE.

  ! calculate the albedo at our starting point

  leaf_transmittance = leaf_single_scattering_albedo_start/ & 
       ( leaf_psd_start+un)
  leaf_reflectance =  leaf_psd_start * &
       leaf_transmittance

  ! some debugging stuff
  !  DO ilevel=1,nlevels_tot
  !
  !     CALL print_debugging_albedo_info(ilevel,cosine_sun_angle,leaf_reflectance,&
  !          leaf_transmittance,&
  !          laieff_collim_pft(ilevel),laieff_isotrop_pft(ilevel), &
  !          reflectance_collim(ilevel-1),reflectance_isotrop(ilevel-1),&
  !          Collim_Alb_Tot_temp(ilevel),Collim_Tran_Tot_temp(ilevel),&
  !          Collim_Abs_Tot_temp(ilevel),&
  !          Isotrop_Alb_Tot_temp(ilevel),Isotrop_Tran_Tot_temp(ilevel),&
  !          Isotrop_Abs_Tot_temp(ilevel))
  !  ENDDO

  ! calculate the gamma coefficients used in the case of the black background 
  call gammas(leaf_reflectance, leaf_transmittance, cosine_sun_angle, gammaCoeffs)
  call gammas(leaf_reflectance,leaf_transmittance,isotropic_cosine_constant,&
       gammaCoeffs_star)



  !************************** step one ****** !
  ! compute all the unscaled quantities
  DO ilevel=nlevels_tot,1,-1

     Collim_Tran_UnColl_Unscaled(ilevel)=exp( - 0.5_r_std*&
          laieff_collim_pft(ilevel)/cosine_sun_angle)

     Isotrop_Tran_UnColl_Unscaled(ilevel)= &
          TBarreUncoll_exact(0.5_r_std*laieff_isotrop_pft(ilevel))


     !	/* 1) collimated source */
     ok=dhrT1(leaf_reflectance,leaf_transmittance,&
          gammaCoeffs(1),gammaCoeffs(2),gammaCoeffs(3),gammaCoeffs(4),&
          sun_zenith_angle_radians, 0.5_r_std*laieff_collim_pft(ilevel),&
          Collim_Alb_Coll_Unscaled(ilevel),Collim_Tran_Coll_Unscaled(ilevel),&
          Collim_Abs_Coll_Unscaled(ilevel))

     ! 	/* 2) isotropic source */
     ok=dhrT1(leaf_reflectance,leaf_transmittance,&
          gammaCoeffs_star(1),gammaCoeffs_star(2),gammaCoeffs_star(3),gammaCoeffs_star(4),&
          acos(isotropic_cosine_constant),0.5_r_std*laieff_isotrop_pft(ilevel),&
          Isotrop_Alb_Coll_Unscaled(ilevel),Isotrop_Tran_Coll_Unscaled(ilevel),&
          Isotrop_Abs_Coll_Unscaled(ilevel))
  ENDDO ! ilevel=nlevels_tot,1,-1


  ! Following the fate of the light at every step

  ! The downwelling array indicates the quantity of light flowing into this level
  ! from above therefore, to start our system for collimated light, we give a 
  ! unit of light coming into the top level from a collimated source.
  ! The upwelling array is light entering this level from below.
  ! cs is the current step, while ns is the next step for the iteration.
  ! Level 0 is the background. Light can enter this level from above but not below
  ! Level nlevels_tot+1 is the atomosphere. Light can enter this level from below 
  ! but not above.
  collim_down_cs(:)=zero
  collim_down_cs(nlevels_tot)=un
  isotrop_down_cs(:)=zero
  isotrop_up_cs(:)=zero

  CALL propagate_fluxes(collim_down_cs, isotrop_down_cs, isotrop_up_cs, &
       Collim_Tran_UnColl_Unscaled, Collim_Tran_Coll_Unscaled, &
       Collim_Alb_Coll_Unscaled, Isotrop_Tran_UnColl_Unscaled, &
       Isotrop_Tran_Coll_Unscaled, Isotrop_Alb_Coll_Unscaled, &
       br_base_temp_collim, br_base_temp_isotrop, .FALSE., &
       Collim_Tran_Uncoll, Collim_Tran_Coll, Collim_Alb_Coll, lconverged, lprint)

  ! Now for the isotropic light
  collim_down_cs(:)=zero
  isotrop_down_cs(:)=zero
  isotrop_down_cs(nlevels_tot)=un
  isotrop_up_cs(:)=zero

  ! The colliminated light is never used here, but it's passed to keep things 
  ! clean between the two cases. Might have to redo this if we are having 
  ! peformance issues.
  CALL propagate_fluxes(collim_down_cs, isotrop_down_cs, isotrop_up_cs, &
       Collim_Tran_UnColl_Unscaled, Collim_Tran_Coll_Unscaled, &
       Collim_Alb_Coll_Unscaled, Isotrop_Tran_UnColl_Unscaled, &
       Isotrop_Tran_Coll_Unscaled, Isotrop_Alb_Coll_Unscaled, &
       br_base_temp_collim, br_base_temp_isotrop, .TRUE., &
       Isotrop_Tran_Uncoll, Isotrop_Tran_Coll, Isotrop_Alb_Coll, lconverged, lprint)

  ! Calculate the absorption profile
  Collim_Tran_Tot(:)=Collim_Tran_Coll(:)+Collim_Tran_Uncoll(:)
  Isotrop_Tran_Tot(:)=Isotrop_Tran_Coll(:)+Isotrop_Tran_Uncoll(:)

  ! bottom level
  Collim_Abs_Tot(1)=Collim_Tran_Tot(1+1)+Collim_Tran_Tot(1)*br_base_temp_collim&
       -Collim_Tran_Tot(1)-Collim_Alb_Coll(1)
  Isotrop_Abs_Tot(1)=Isotrop_Tran_Tot(1+1)+Isotrop_Tran_Tot(1)*br_base_temp_isotrop&
       -Isotrop_Tran_Tot(1)-Isotrop_Alb_Coll(1)

  ! all middle levels
  DO ilevel=2,nlevels_tot-1
     Collim_Abs_Tot(ilevel)=Collim_Tran_Tot(ilevel+1)+Collim_Alb_Coll(ilevel-1)&
          -Collim_Tran_Tot(ilevel)-Collim_Alb_Coll(ilevel)
     Isotrop_Abs_Tot(ilevel)=Isotrop_Tran_Tot(ilevel+1)+Isotrop_Alb_Coll(ilevel-1)&
          -Isotrop_Tran_Tot(ilevel)-Isotrop_Alb_Coll(ilevel)
  ENDDO

  ! top level
  Collim_Abs_Tot(nlevels_tot)=un+Collim_Alb_Coll(nlevels_tot-1)&
       -Collim_Tran_Tot(nlevels_tot)-Collim_Alb_Coll(nlevels_tot)
  Isotrop_Abs_Tot(nlevels_tot)=un+Isotrop_Alb_Coll(nlevels_tot-1)&
       -Isotrop_Tran_Tot(nlevels_tot)-Isotrop_Alb_Coll(nlevels_tot)

END SUBROUTINE multilevel_albedo

!! ==============================================================================================================================
!! SUBROUTINE   : print_debugging_albedo_info
!!
!>\BRIEF         Prints out some albedo information in a nice format.  
!!               Should only be used for debugging, never for production runs.
!!
!! DESCRIPTION : 
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): None.
!!
!! REFERENCE(S) :  
!!
!! FLOWCHART    : None
!_ ================================================================================================================================
SUBROUTINE print_debugging_albedo_info(ilevel, cosine_sun_angle, &
     leaf_reflectance, leaf_transmittance, laieff_collim_temp, laieff_isotrop_temp, &
     br_base_temp_collim, br_base_temp_isotrop, Collim_Alb_Tot, &
     Collim_Tran_Tot, Collim_Abs_Tot, Isotrop_Alb_Tot, Isotrop_Tran_Tot, Isotrop_Abs_Tot)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  INTEGER,INTENT(IN)          :: ilevel
  REAL(r_std), INTENT(IN)     :: cosine_sun_angle            !! cosine of the solar zenith angle, between 0 and 1
                                                             !! @tex $()$ @endtex 
   REAL(r_std), INTENT(IN)    :: laieff_collim_temp           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: laieff_isotrop_temp           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: leaf_reflectance           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: leaf_transmittance            !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: br_base_temp_collim           !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std), INTENT(IN)    :: br_base_temp_isotrop            !! cosine of the solar zenith angle, between 0 and 1
   REAL(r_std),INTENT(IN)     :: Collim_Alb_Tot
   REAL(r_std),INTENT(IN)     :: Collim_Tran_Tot
   REAL(r_std),INTENT(IN)     :: Collim_Abs_Tot
   REAL(r_std),INTENT(IN)     :: Isotrop_Alb_Tot
   REAL(r_std),INTENT(IN)     :: Isotrop_Tran_Tot
   REAL(r_std),INTENT(IN)     :: Isotrop_Abs_Tot

  !! 0.2 Output variables

  !! 0.3 Modified variables

  !! 0.4 Local variables

  !_ ================================================================================================================================
   
1223 FORMAT(I5,I3,7(F14.7))

     WRITE (numout,'(8(11A))') '   Level   ','   Angle   ','  laieff_c  ','  laieff_i  ',&
          '   rleaf   ','   tleaf   ','   rbgd_c   ','   rbgd_i   '

     WRITE(numout,FMT='(I6,5X,7(F11.6))') &
          ilevel,180/pi*ACOS(cosine_sun_angle),&
          laieff_collim_temp,laieff_isotrop_temp,&
          leaf_reflectance,leaf_transmittance,br_base_temp_collim,&
          br_base_temp_isotrop


     WRITE (numout,'(10X,6(11A))') ' Rtot(sun) ',' Ttot(sun) ',&
          ' Atot(sun) ',' Rtot(iso) ',' Ttot(iso) ',' Atot(iso) '
     WRITE(numout,FMT='(10X,6(F11.6))') &
          Collim_Alb_Tot,Collim_Tran_Tot,Collim_Abs_Tot,&
          Isotrop_Alb_Tot,Isotrop_Tran_Tot,Isotrop_Abs_Tot


END SUBROUTINE print_debugging_albedo_info


!! ==============================================================================================================================
!! SUBROUTINE   : propagate_fluxes
!!
!>\BRIEF         Propogates the radiation fluxes through each level of the
!!               canopy.
!!
!! DESCRIPTION : This is an algorithm to follow radition fluxes through the
!!     canopy.  At each level, the path probabilities are determined by the
!!     raditional transfer scheme of Pinty et al (2006).  Notice that
!!     the fluxes given by this routine are all cumulative fluxes, not per
!!     level.
!!
!! RECENT CHANGE(S): None\n
!!
!! MAIN OUTPUT VARIABLE(S): ::Tran_Uncoll_Tot, ::Tran_Coll_Tot, ::Alb_Coll_Tot
!!
!! REFERENCE(S) :  
!!
!! FLOWCHART    : None
!_ ================================================================================================================================
SUBROUTINE propagate_fluxes(collim_down_cs, isotrop_down_cs, isotrop_up_cs, &
     Collim_Tran_UnColl_Unscaled, Collim_Tran_Coll_Unscaled, &
     Collim_Alb_Coll_Unscaled, Isotrop_Tran_UnColl_Unscaled, &
     Isotrop_Tran_Coll_Unscaled, Isotrop_Alb_Coll_Unscaled, br_base_temp_collim, &
     br_base_temp_isotrop, lisotrop, Tran_Uncoll_Tot, Tran_Coll_Tot, &
     Alb_Coll_Tot, lconverged, lprint)

  !! 0. Variables and parameter declaration
  !! 0.1 Input variables
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(IN)           :: Collim_Tran_UnColl_Unscaled
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(IN)           :: Collim_Tran_Coll_Unscaled
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(IN)           :: Collim_Alb_Coll_Unscaled
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(IN)           :: Isotrop_Tran_UnColl_Unscaled
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(IN)           :: Isotrop_Tran_Coll_Unscaled
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(IN)           :: Isotrop_Alb_Coll_Unscaled
  REAL(r_std), INTENT(IN)                                  :: br_base_temp_collim             !! cosine of the solar zenith angle, between 0 and 1
  REAL(r_std), INTENT(IN)                                  :: br_base_temp_isotrop            !! cosine of the solar zenith angle, between 0 and 1
  LOGICAL, INTENT(IN)                                      :: lisotrop                        !! are we dealing with an isotropic source?  only needed for correct partitioning
                                                                                              !! of the collided and uncollided transmitted light...the total is not affected
  LOGICAL, INTENT(IN)                                      :: lprint                          !! a flag to print

  !! 0.2 Output variables
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(OUT)          :: Tran_Uncoll_Tot
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(OUT)          :: Tran_Coll_Tot
  REAL(r_std), DIMENSION(nlevels_tot),INTENT(OUT)          :: Alb_Coll_Tot
  LOGICAL,INTENT(OUT)                                      :: lconverged

  !! 0.3 Modified variables
  REAL(r_std), DIMENSION(0:nlevels_tot+1),INTENT(INOUT)    :: collim_down_cs
  REAL(r_std), DIMENSION(0:nlevels_tot+1),INTENT(INOUT)    :: isotrop_down_cs
  REAL(r_std), DIMENSION(0:nlevels_tot+1),INTENT(INOUT)    :: isotrop_up_cs

  !! 0.4 Local variables
  INTEGER :: istep,ilevel
  REAL(r_std),DIMENSION(0:nlevels_tot+1)            :: collim_down_ns,isotrop_down_ns,&
       isotrop_up_ns,Tran_Uncoll_Tot_temp
  
  !_ ================================================================================================================================

  Tran_Uncoll_Tot(:)=zero
  Tran_Coll_Tot(:)=zero
  Alb_Coll_Tot(:)=zero

  Tran_Uncoll_Tot_temp(:)=un

  istep=0

  DO

     istep=istep+1

     lconverged=.TRUE.

     ! Zero out the counters for the next step
     collim_down_ns(:)=zero
     isotrop_down_ns(:)=zero
     isotrop_up_ns(:)=zero


     ! Now we need to loop over all levels and see what light is entering the 
     ! level, and how it will propagate in the next step
     DO ilevel=1,nlevels_tot

        ! For collimated downwelling light into the level, it can be scattered 
        ! up, down, or pass through uncollided
        IF(collim_down_cs(ilevel) .GT. zero)THEN
           collim_down_ns(ilevel-1)=collim_down_ns(ilevel-1)+&
                collim_down_cs(ilevel)*Collim_Tran_UnColl_Unscaled(ilevel)


           ! This statement checks to see if this light has been previously
           ! scattered or not.  This term is only present in level nlevels_tot
           ! for the first step, level nlevels_tot-1 for the second step, 
           ! level nlevels_tot-2 for the third step, etc., and it only happens
           ! in the case of a collimated light source.
           IF(ilevel == nlevels_tot-istep+1 .AND. .NOT. lisotrop) THEN
              Tran_Uncoll_Tot_temp(ilevel)=Tran_Uncoll_Tot_temp(ilevel+1)*&
                   Collim_Tran_UnColl_Unscaled(ilevel)
           ENDIF

           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                collim_down_cs(ilevel)*Collim_Tran_Coll_Unscaled(ilevel)
           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+&
                collim_down_cs(ilevel)*Collim_Alb_Coll_Unscaled(ilevel)

        ENDIF ! collim_down_cs(ilevel) .GT. zero

        ! For isotropic downwelling light, it can also be scattered up, down,
        ! or pass through uncollided
        IF(isotrop_down_cs(ilevel) .GT. zero)THEN
           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                isotrop_down_cs(ilevel)*Isotrop_Tran_UnColl_Unscaled(ilevel)

           ! This is the same check as above, but this time the light has
           ! an isotropic source and not a collimated source
           IF(ilevel == nlevels_tot-istep+1 .AND. lisotrop) THEN
              Tran_Uncoll_Tot_temp(ilevel)=Tran_Uncoll_Tot_temp(ilevel+1)&
                   *Isotrop_Tran_UnColl_Unscaled(ilevel)
           ENDIF

           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                isotrop_down_cs(ilevel)*Isotrop_Tran_Coll_Unscaled(ilevel)
           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+&
                isotrop_down_cs(ilevel)*Isotrop_Alb_Coll_Unscaled(ilevel)
        ENDIF ! isotrop_down_cs(ilevel) .GT. zero

        ! Isotropic upwelling light can pass through upwards collided or 
        ! uncollided with vegetation in this level, or it can be reflected downwards
        IF(isotrop_up_cs(ilevel) .GT. zero)THEN

           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+isotrop_up_cs(ilevel)&
                *Isotrop_Tran_UnColl_Unscaled(ilevel)
           isotrop_up_ns(ilevel+1)=isotrop_up_ns(ilevel+1)+isotrop_up_cs(ilevel)&
                *Isotrop_Tran_Coll_Unscaled(ilevel)
           isotrop_down_ns(ilevel-1)=isotrop_down_ns(ilevel-1)+&
                isotrop_up_cs(ilevel)*Isotrop_Alb_Coll_Unscaled(ilevel)
        ENDIF

        ! there can be no collimated upwards light, since upwards light must 
        ! always have reflected off something


     ENDDO ! ilevel=1,nlevels_tot

     ! The background is a bit special. there is no transmission, but there is 
     ! reflection, which leads to a source term to the bottom level from below.
     isotrop_up_ns(1)=isotrop_up_ns(1)+collim_down_cs(0)*br_base_temp_collim
     isotrop_up_ns(1)=isotrop_up_ns(1)+isotrop_down_cs(0)*br_base_temp_isotrop


     ! Now we add the light generated here to our cumulative counters to track
     ! the total amount that leaves the canopy, either through being absorbed
     ! by the background or being reflected back into the atmosphere.
     ! We keep track of the uncoll above, so here we track the total light that
     ! is transmitted to the soil, and then at the end we taken the difference 
     ! to get the collided radation.
     Tran_Coll_Tot(1:nlevels_tot)=Tran_Coll_Tot(1:nlevels_tot)+&
          isotrop_down_ns(0:nlevels_tot-1)+collim_down_ns(0:nlevels_tot-1)
     Alb_Coll_Tot(1:nlevels_tot)=Alb_Coll_Tot(1:nlevels_tot)+isotrop_up_ns(2:nlevels_tot+1)

     ! now we update the light we are currently tracking
     collim_down_cs(:)=collim_down_ns(:)
     isotrop_down_cs(:)=isotrop_down_ns(:)
     isotrop_up_cs(:)=isotrop_up_ns(:)


     ! check for convegence...if all the values of light are currently below a 
     ! threshold, we're probably good assuming the threshold is low enough.
     IF(MAXVAL(collim_down_cs,1) .GT. alb_threshold) lconverged=.FALSE.
     IF(MAXVAL(isotrop_down_cs,1) .GT. alb_threshold) lconverged=.FALSE.
     IF(MAXVAL(isotrop_up_cs,1) .GT. alb_threshold) lconverged=.FALSE.

     IF(lconverged)THEN
        IF(lprint) WRITE(numout,*) 'Converged after this many steps: ',istep
        EXIT
     ENDIF

     ! This number here could also be externalized.
     IF(istep .GE. 1000)THEN
        WRITE(numout,*) '*********************************************************'
        WRITE(numout,'(A,I6,F14.10)') 'Albedo not converging!',istep,alb_threshold
        WRITE(numout,'(A)') '                  collim_down_cs ' // &
             'isotrop_down_cs  isotrop_up_cs'
        DO ilevel=0,nlevels_tot+1
           WRITE(numout,'(I4,3F14.10)') ilevel, &
                collim_down_cs(ilevel),isotrop_down_cs(ilevel),isotrop_up_cs(ilevel)

        ENDDO
        WRITE(numout,*) 'You should increase either the number' // &
             'of steps or the alb_threshold.'
        WRITE(numout,*) '*********************************************************'
        EXIT
     ENDIF ! istep .GE. 1000
  ENDDO ! convergence loop


  ! now separate the collided from the uncollided light.  Notice that
  ! this is not really needed for any purposes other than debugging, as the
  ! important quantity is the total amount of light striking the ground.
  Tran_UnColl_Tot(1:nlevels_tot)=Tran_UnColl_Tot_temp(1:nlevels_tot)
  Tran_Coll_Tot(1:nlevels_tot)=Tran_Coll_Tot(1:nlevels_tot)-Tran_UnColl_Tot(1:nlevels_tot)

  ! Some debugging information
  IF(lprint)THEN
     WRITE(numout,'(A)') '      Tran_Uncoll_Tot   Tran_Coll_Tot  Alb_Coll_Tot'
     DO ilevel=nlevels_tot,1,-1
        WRITE(numout,'(I4,3F10.6)') ilevel, &
             Tran_Uncoll_Tot(ilevel),Tran_Coll_Tot(ilevel),Alb_Coll_Tot(ilevel)
     ENDDO
  ENDIF

END SUBROUTINE propagate_fluxes


! ----------------------------------------------------------------------------------------


END MODULE albedo
